/***********************************************************************
 * lut.c
 *
 * Copyright (C) 1997-2002 Klaus Ehrenfried
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * 
 * Klaus Ehrenfried 26-Mar-97
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>

#include "main.h"
#include "lut.h"

#define MAXLINE 512
#define MAXCTLEN 512

/************************************************************************
 *   Lese Zeile
 ***********************************************************************/

static int getline(FILE *fp, char s[], int len)
{
  int i,c;

  for (i=0; i < len-1; i++)
    {
      c=getc(fp);
      if (c == EOF)
	{
	  s[i]='\0';
	  return(1);
	}
      if (c == '\n')
	{
	  s[i]='\0';
	  return(0);
	}
      s[i]=c;
    }
  s[len-1]='\0';
  return(0);
}

/************************************************************************
 *   Suche ersten Character in Record
 ***********************************************************************/

static char *getstart(char *s, int len)
{
  int i;

  for (i=0; i < len; i++)
    {
      if ((*s != ' ') && (*s != '\t')) break;
      s++;
    }

  return(s);
}

/************************************************************************
 * load_lut
 ************************************************************************/

void load_lut(char *name, TASK *a)
{
   FILE *fp;
   LUT_ENTRY *plut;
   UBYTE *pt;
   char rbuf[MAXLINE];
   char *b;
   double fa[3],fb[3];
   double *f1,*f2;
   double wa,wb;
   double deps,eps,emeps;
   double wmin,wmax;
   int count,ii,k,n,i,j;
   int color;

   fp=fopen(name,"r");
   if (fp == NULL)
     {
       fprintf(stderr,"Error opening lut-file '%s'\n",name);
       exit(1);
    }

   plut=a->lut;
   for (n=0; n < a->lut_count; n++) {
     if (plut->table != NULL) free(plut->table);
     plut++;
   }

   count=0;
   plut=a->lut;
   wmin=0.0;
   wmax=1.0;
   while (count < a->lut_max_entries)
     {
	k=getline(fp, rbuf, MAXLINE);
	b=getstart(rbuf, MAXLINE);
	if (*b != '#')
	  {
	     ii=sscanf(rbuf,"%lf %lf,%lf,%lf %lf %lf,%lf,%lf %d",
		       &wa,fa,fa+1,fa+2,&wb,fb,fb+1,fb+2,&n);
	     if (ii == 9)
	       {
		  if (wa < wb)
		    {
		       plut->w1=wa;
		       plut->w2=wb;
		       f1=fa;
		       f2=fb;
		    }
		  else
		    {
		       plut->w1=wb;
		       plut->w2=wa;
		       f1=fb;
		       f2=fa;
		    }

		  if (count == 0) {
		    wmin=plut->w1;
		    wmax=plut->w2;
		  }
		  else {
		    if (plut->w1 < wmin) wmin=plut->w1;
		    if (plut->w2 > wmax) wmax=plut->w2;
		  }

		  if (n < 0) n=2;
		  if (n > MAXCTLEN) n=MAXCTLEN;

		  pt=(char *)malloc(3*n);
		  if (pt == NULL) {
		    printf("No memory for color table (%d)\n",n);
		    exit(1);
		  }
		  plut->table=pt;
		  deps=1.0/(double)(n-1);

		  for (i=0; i < n; i++) {
		    eps = i * deps;
		    emeps = 1.0-eps;

		    for (j=0; j < 3; j++)
		      {
			color = 256*(emeps * f1[j] + eps * f2[j]) - 0.5;
			if (color > 255) color=255;
			if (color < 0) color=0;
			*pt++ = color;
		      }
		  }

		  plut->max=n-1;
		  plut->step=(plut->w2 - plut->w1)/(double)n;
		  if (plut->step <= 0.0) plut->step=1.0;

		  plut++;
		  count++;
	       }
	     else if (*b != '\0')
	       {
		  printf("Invalid lut entry: '%s'\n",rbuf);
	       }
	  }
	if (k == 1) break;
     }
   
   a->lut_count = count;
   if (count >= a->lut_max_entries)
     {
	printf("Probably too many lut entries (> %d)\n",a->lut_max_entries);
     }
   
   a->lut_wmin=wmin;
   a->lut_wmax=wmax;
   a->lut_dw=(wmax-wmin);
   if (a->lut_dw <= 0.0) a->lut_dw=1.0;

   printf("lut: tables=%d  wmin=%f wmax=%f\n",count,wmin,wmax);

   fclose(fp);
}

/************************************************************************
 * look_up_color
 ************************************************************************/

void look_up_color(double w, int c[3], TASK *a)
{
   LUT_ENTRY *p;
   UBYTE *pt;
   double h;
   int m,n;

   /*printf("lu (w): %f\n",w);*/

   if (a->lut_periodic == 1) {
     if (w < a->lut_wmin) {
       h = (a->lut_wmax - w)/a->lut_dw;
       w = a->lut_wmax - a->lut_dw * (h - floor(h));
     }
     if (w > a->lut_wmax) {
       h = (w - a->lut_wmin)/a->lut_dw;
       w = a->lut_wmin + a->lut_dw * (h - floor(h));
     }
   }

   c[0]=0;
   c[1]=0;
   c[2]=0;

   p=a->lut;
   for (n=0; n < a->lut_count; n++)
     {
	if ((w >= p->w1) && (w <= p->w2))
	  {
	     h = (w - p->w1)/p->step;
	     m = (int)h;
	     if (m > p->max) m=p->max;
	     pt = p->table + 3*m;

	     c[0]=pt[0];
	     c[1]=pt[1];
	     c[2]=pt[2];
	  }
	p++;
     }
}
