/***********************************************************************
 * main.c
 * 
 * Copyright (C) 1997-2002 Klaus Ehrenfried
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * 
 * Klaus Ehrenfried 26-Mar-97
 * 
 * 7-Mar-2000 added hidden_color K.E.
 * 30-Dec-2001 added out_tolerance and shift K.E.
 * 4-Jan-2002 improved set_cont_levels K.E.
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <math.h>

#define MAXLINE 512
#define MAXDIGITS 16

#define YESNO(A) switch (b3[0])\
{case 'n': A=0; break; case 'y': A=1; break; default: return(1);}

#include "main.h"
#include "lut.h"
#include "object.h"
#include "draw3d.h"
#include "text.h"

#define MAXPAR 10

typedef struct param_struct
{
   char *v[MAXPAR];
   int flag[MAXPAR];
   int count;
} PARAM;

static const char inifile[]=".pixmaxrc";
static const char ppmfile_name[]="PIXMAX.PPM";

#ifdef FONTPATH
static const char fontpath_name[]=FONTPATH;
#else
static const char fontpath_name[]="/usr/local/lib/pixmax";
#endif


/************************************************************************
 *   eval_param
 ***********************************************************************/

static void eval_param(PARAM *param, char *s)
{
   int i,n,len;
   int active;
   char *b;

   len=strlen(s);
   for (i=0; i < len; i++)
     {
	if (s[i] == '\0') s[i] = ' ';
	if (s[i] == '\t') s[i] = ' ';
     }

   for (i=1; i < len; i++)
     {
	if ((s[i] == ' ') && ((s[i-1] == ',') || (s[i-1] == '\t'))) {
	   s[i] = '\t';
	}
     }

   for (i=len-2; i >= 0; i--)
     {
	if ((s[i] == ' ') && ((s[i+1] == ',') || (s[i+1] == '\t'))) {
	   s[i] = '\t';
	}
     }

   n=0;
   active=0;

   for (i=0; i < len; i++)
     {
	if (active == 0)
	  {
	     if (s[i] != ' ')
	       {
		  param->v[n++]=s+i;
		  active=1;
		  if (s[i] == '[')
		    {active=2;}
		  else if (s[i] == '"')
		    {active=3;}
	       }
	  }
	else if (active == 1)
	  {
	     if (s[i] == ' ')
	       {
		  active=0;
		  s[i]='\0';
	       }
	  }
	else if (active == 2)
	  {
	     if (s[i] == ']')
	       {
		  active=1;
	       }
	  }
	else if (active == 3)
	  {
	     if (s[i] == '"')
	       {
		  active=1;
	       }
	  }
	
	if (n >= MAXPAR) break;
     }

   for (i=1; i < len; i++)
     {
	if (s[i] == '\t') {
	   s[i] = ' ';
	}
     }

   param->count=n;

   for (n=0; n < param->count; n++)
     {
	b=param->v[n];
	param->flag[n]=0;
	if (b[0] == '"')
	  {
	     len=strlen(b);
	     if ((len > 1) && (b[len-1] == '"'))
	       {
		  b[len-1]='\0';
		  (param->v[n])++;
		  param->flag[n]=1;
	       }
	  }
	else if (b[0] == '[')
	  {
	     len=strlen(b);
	     if ((len > 1) && (b[len-1] == ']'))
	       {
		  b[len-1]='\0';
		  (param->v[n])++;
		  param->flag[n]=2;
	       }
	  }
     }
}

/************************************************************************
 *   Lese Zeile
 ***********************************************************************/

static int getline(FILE *fp, char s[], int len)
{
   int i,c;

   for (i=0; i < len-1; i++)
     {
	c=getc(fp);
	if (c == EOF)
	  {
	     s[i]='\0';
	     return(1);
	  }
	if (c == '\n')
	  {
	     s[i]='\0';
	     return(0);
	  }
	if (c == '\t') c=' ';
	s[i]=c;
     }
   s[len-1]='\0';
   return(0);
}

/************************************************************************
 * init_task
 ************************************************************************/

static void init_task(TASK *a)
{
   int j,n;

   printf("Image: %dx%d\n",a->image_width,a->image_height);
   printf("line_jump_max: %d\n",a->line_jump_max);
   printf("lut_max_entries: %d\n",a->lut_max_entries);

   a->image_len=a->image_width * a->image_height;

   if ((a->image_width <= 0) ||
       (a->image_height <= 0))
     {
      fprintf(stderr,"Stupid %dx%d image\n",
	      a->image_width,a->image_height);
      exit(1);
   }

   a->image_center[0]=0.5*a->image_width;
   a->image_center[1]=0.5*a->image_height;

   a->pels=(PIXEL *)calloc(a->image_len, sizeof(PIXEL));

   if (a->pels == NULL) {
      fprintf(stderr,"No memory for %dx%d image\n",
	      a->image_width,a->image_height);
      exit(1);
   }

   a->line_jump=(int *)calloc(2*a->line_jump_max, sizeof(int));

   if (a->line_jump == NULL) {
      fprintf(stderr,"No memory for line-jump table (%d)\n",a->line_jump_max);
      exit(1);
   }
   a->line_jump_len=0;

   a->lut=(LUT_ENTRY *)calloc(a->lut_max_entries, sizeof(LUT_ENTRY));

   if (a->lut == NULL) {
      fprintf(stderr,"No memory for lut (%d)\n",a->lut_max_entries);
      exit(1);
   }
   a->lut_count=0;
   a->lut_periodic=0;

   a->view_alpha=30.0;
   a->view_beta=30.0;
   a->view_zoom=1.0;
   a->view_dist=10.0;
   init_tmatrix(a);

   strcpy(a->axis_format,"%.1f");

   a->out_tolerance=128;
   a->line_dz=0.02;
   /*a->line_check_dz=0.01;*/
   a->line_check_width=1;
   a->line_pixel_shift=0;
   a->line_width=0;
   a->edge_width=0;
   a->border_width=2;

   a->view_box[0]=1.0;
   a->view_box[1]=1.0;
   a->view_box[2]=1.0;
   
   a->view_auto_scale[0]=1;
   a->view_auto_scale[1]=1;
   a->view_auto_scale[2]=1;

   for (j=0; j < 3; j++) {
      a->line_color[j]=255;
      a->edge_color[j]=255;
      a->face_color[j]=255;
      a->hidden_color[j]=127;
      a->border_color[j]=255;
      a->text_color_bg[j]=0;
      a->text_color_fg[j]=255;
      a->render_hidden_color[j]=0.5;
   }

   a->cont_levels=0;

   a->object_cont=0;
   a->object_edges=1;
   a->object_faces=0;
   a->object_pos[0]=1;
   a->object_pos[1]=2;
   a->object_pos[2]=3;
   a->object_pos[3]=1;

   a->grid_faces=1;
   a->grid_border=0;

   a->line_shading=0;
   a->face_shading=0;
   a->edge_shading=0;
   a->draw_clu_flag=0;

   a->render_ambient=0.2;
   a->render_color[0]=1.0;
   a->render_color[1]=1.0;
   a->render_color[2]=1.0;

   for (n=0; n < MAXLIGHTS; n++) {
      a->light[n].alpha=0.0;
      a->light[n].beta=0.0;
      a->light[n].flag=0;
      a->light[n].power=0.0;
   }
   a->light[0].power=0.6;

   a->font_pels=NULL;
   a->font_width=0;
   a->font_height=0;

   a->text_center[0]=0;
   a->text_center[1]=0;
   a->text1_point.x=0.0;
   a->text1_point.y=0.0;
   a->text1_point.z=0.0;
   a->text1_point.w=0.0;

   a->axis_ticks[0]=3;
   a->axis_ticks[1]=3;
   a->axis_ticks[2]=3;

   a->axis_tick_len=1.05;
   a->axis_tick_jump=1.1;

   a->colorbar_pos[0]=20;
   a->colorbar_pos[1]=20;

   a->colorbar_size[0]=200;
   a->colorbar_size[1]=20;

   a->colorbar_skip[0]=5;
   a->colorbar_skip[1]=10;

   a->colorbar_ticks=3;
}

/************************************************************************
 * set_null
 ************************************************************************/

int set_null(TASK *a)
{
   int j;

   a->object_cont=0;
   a->object_edges=1;
   a->object_faces=1;
   a->object_pos[0]=1;
   a->object_pos[1]=2;
   a->object_pos[2]=3;
   a->object_pos[3]=1;

   a->grid_faces=1;
   a->grid_border=0;

   a->line_shading=0;
   a->face_shading=0;
   a->edge_shading=0;
   a->draw_clu_flag=0;

   for (j=0; j < 3; j++) {
      a->line_color[j]=0;
      a->edge_color[j]=0;
      a->face_color[j]=255;
      a->border_color[j]=255;
   }

   a->line_dz=0.02;
   a->line_width=1;
   a->edge_width=1;
   a->border_width=2;

   a->out_tolerance=128;

   return(0);
}

/************************************************************************
 * write_ppm
 ************************************************************************/

void write_ppm(TASK *a)
{
  FILE *fp;
  PIXEL *pp;
  int i;

  fp=fopen(a->ppmfile,"w");
  if (fp == NULL)
    {
      fprintf(stderr,"Error opening file '%s'\n",a->ppmfile);
      exit(1);
    }

  fprintf(fp,"P6\n");
  fprintf(fp,"# Created by rasterman\n");
  fprintf(fp,"%d %d\n",a->image_width,a->image_height);
  fprintf(fp,"255\n");

  pp=a->pels;
  for (i=0; i < a->image_len; i++)
    {
      fputc(pp->r,fp);
      fputc(pp->g,fp);
      fputc(pp->b,fp);
      pp++;
    }

  fclose(fp);
}

/************************************************************************
 * draw3_off
 ************************************************************************/

static void draw3_off(char *fname, TASK *a)
{
  FACE_OBJECT obj;

  obj.pv=NULL;
  obj.pf=NULL;

  if (read_off(fname, &obj, a->object_pos) == 0) {
    draw2_face_object(&obj, a);
  }
  else {
    printf("read_off failed!\n");
    exit(1);
  }

  printf("O.K.\n");
  if (obj.pv != NULL) free(obj.pv);
  if (obj.pf != NULL) free(obj.pf);
}

/************************************************************************
 * draw3_mesh
 ************************************************************************/

static void draw3_mesh(char *fname, TASK *a)
{
  MESH_OBJECT mesh;

  mesh.pv=NULL;
  mesh.id=NULL;

  if (read_mesh(fname, &mesh, a->object_pos) == 0) {
    draw2_mesh_object(&mesh, a);
  }
  else {
    printf("read_mesh failed!\n");
    exit(1);
  }

  printf("O.K.\n");
  if (mesh.pv != NULL) free(mesh.pv);
  if (mesh.id != NULL) free(mesh.id);
}

/************************************************************************
 * background
 ************************************************************************/

static int background(PARAM *param, TASK *a)
{
   PIXEL *pp;
   int red,green,blue;
   int i;

   if (param->count < 2) return(1);
   if (sscanf(param->v[1],"%d,%d,%d",&red,&green,&blue) != 3) return(1);

   pp=a->pels;
   for (i=0; i < a->image_len; i++)
     {
       if (pp->hit == 0)
	 {
	   pp->r=red;
	   pp->g=green;
	   pp->b=blue;
	 }
       pp++;
     }

   return(0);
}

/************************************************************************
 * set_appear
 ************************************************************************/

void set_appear(char *s, TASK *a)
{
  int flag;

  flag=0;

  a->object_edges=0;
  a->object_faces=0;
  a->object_cont=0;
  a->edge_shading=0;
  a->face_shading=0;
  a->line_shading=0;

  while (*s != '\0')
    {
      switch (*s)
	{
	case 'F':
	case 'f': a->object_faces=1; flag=1; break;
	case 'E':
	case 'e': a->object_edges=1; flag=2; break;
	case 'C':
	case 'c': a->object_cont=1; flag=3; break;
	case 'S':
	case 's':
	  if (flag == 1) {
	    a->face_shading=1;
	  }
	  else if (flag == 2) {
	    a->edge_shading=1;
	  }
	  else if (flag == 3) {
	    a->line_shading=1;
	  }
	  flag=0;
	  break;
	case 'R':
	case 'r':
	  if (flag == 1) {
	    a->face_shading=2;
	  }
	  flag=0;
	  break;
	default: flag=0; break;
	}
      s++;
    }
}

/************************************************************************
 * set_cont_levels
 ************************************************************************/

void set_cont_levels(double d1, double d2, double d3, TASK *a)
{
   int n;
   int count;
   double dx;

   d2 = fabs(d2);

   count=0;
   if (d3 >= d1) {
      for (n=0; n < MAXCONTLEV; n++)
	{
	   dx = d1 + n*d2;
	   if (dx > d3) break;
	   a->cont_value[count++] = dx;
	}
   } else {
      for (n=0; n < MAXCONTLEV; n++)
	{
	   dx = d1 - n*d2;
	   if (dx < d3) break;
	   a->cont_value[count++] = dx;
	}
   }
   a->cont_levels=count;
   printf("%d levels\n",count);
}

/************************************************************************
 * set_view_box
 ************************************************************************/

void set_view_box(double d1, double d2, double d3, TASK *a)
{
   double dmax;

   if (d1 < 0.0) d1=-d1;
   if (d2 < 0.0) d2=-d2;
   if (d3 < 0.0) d3=-d3;

   if (d1 == 0.0) d1=1.0;
   if (d2 == 0.0) d2=1.0;
   if (d3 == 0.0) d3=1.0;

   dmax=d1;
   if (d2 > dmax) dmax=d2;
   if (d3 > dmax) dmax=d3;

   if (dmax > 0) {
      d1 /= dmax;
      d2 /= dmax;
      d3 /= dmax;
   }
   a->view_box[0]=d1;
   a->view_box[1]=d2;
   a->view_box[2]=d3;
}

/************************************************************************
 * set_axis_format
 ************************************************************************/

static void set_axis_format(char *fmt, TASK *a)
{
  int len,pos1,ext,digits;
  int n,k;

  if (fmt[0] == '%') fmt++;
  len=strlen(fmt);
  if (len > MAXLINE) len=MAXLINE;

  pos1=-1;
  ext='f';

  for (n=0; n < len; n++)
    {
      switch (fmt[n])
	{
	case '.':
	  pos1=n;
	  break;
	case 'e':
	case 'E':
	  ext=fmt[n];
	  break;
	default:
	  break;
	}
    }

  if (pos1 == -1) {
    digits=0;
  }
  else {
    k=sscanf(fmt+pos1+1,"%d",&digits);
    if (k != 1) digits=0;
    if (digits > MAXDIGITS) digits=MAXDIGITS;
    if (digits < 0) digits=0;
  }

  switch (ext)
    {
    case 'e':
      sprintf(a->axis_format,"%%.%de",digits);
      break;
    case 'E':
      sprintf(a->axis_format,"%%.%dE",digits);
      break;
    default:
      sprintf(a->axis_format,"%%.%df",digits);
      break;
    }
}

/************************************************************************
 * test_token
 ************************************************************************/

int test_token(char *s1, char *s2)
{
   int len,i;

   len=strlen(s1);

   for (i=0; i < len; i++)
     {
	if (s1[i] != s2[i]) return(0);
     }
   return(1);
}

/************************************************************************
 * command_axis
 ************************************************************************/

int command_axis(PARAM *param, int dir, TASK *a)
{
  char *b3;
  double d1,d2;
  double h[2];

  if (param->count < 2) return(1);

  if (sscanf(param->v[1],"%lf,%lf",&d1,&d2) != 2) return(2);

  h[0]=d1;
  h[1]=d2;

  a->text_center[0]=0;
  a->text_center[1]=0;

  if (param->count > 2)
    {
      b3=param->v[2];
      if (b3[0] == 'l')
	{
	  a->text_center[0]=-1;
	}
      else if (b3[0] == 'r')
	{
	  a->text_center[0]=1;
	}
      if (b3[1] == 't')
	{
	  a->text_center[1]=-1;
	}
      else if (b3[1] == 'b')
	{
	  a->text_center[1]=1;
	}
    }
  draw1_axis(dir,h,a);
  return(0);
}

/************************************************************************
 * command_text
 ************************************************************************/

int command_text(PARAM *param, TASK *a)
{
   RPOINT rp;
   int n1,n2;

   if (param->count < 3) return(1);
   if (sscanf(param->v[1],"%d,%d",&n1,&n2) != 2) return(2);

   if (n1 >= 0)
     {rp.i=n1;}
   else
     {rp.i=a->image_width-n1;}

   if (n2 >= 0)
     {rp.j=n2;}
   else
     {rp.j=a->image_height-n2;}

   rp.z=0.0;
   rp.w=0.0;
   draw0_text(&rp,param->v[2],a);

  return(0);
}

/************************************************************************
 * command_3dtext
 ************************************************************************/

int command_3dtext(PARAM *param, TASK *a)
{
  POINT p;
  double d1,d2,d3;

  if (param->count < 3) return(1);
  if (sscanf(param->v[1],"%lf,%lf,%lf",&d1,&d2,&d3) != 3) return(2);

  p.x=d1 * a->view_box[0];
  p.y=d2 * a->view_box[1];
  p.z=d3 * a->view_box[2];

  draw1_text(&p,param->v[2],a);

  return(0);
}

/************************************************************************
 * command_draw
 ************************************************************************/

int command_draw(PARAM *param, TASK *a)
{
   char *b2,*b3,*b4;
   double d1,d2;
   int n1,n2;

   if (param->count < 2) return(2);
   b2=param->v[1];

   if (test_token("box",b2) == 1)
     {
       draw1_view_box(a);
     }
   else
     {
       if (param->count < 3) return(3);
       b3=param->v[2];
       if (test_token("obj",b2) == 1)
	 {
	   draw3_off(b3, a);
	 }
       else if (test_token("mesh",b2) == 1)
	 {
	   draw3_mesh(b3, a);
	 }
	else if (test_token("color",b2) == 1)
	  {
	     if (sscanf(b3,"%lf,%lf",&d1,&d2) != 2) return(3);
	     draw0_colorbar(d1,d2,a);
	  }
       else
	 {
	   if (param->count < 4) return(4);
	   b4=param->v[3];
	   if (test_token("yz_grid",b2) == 1)
	     {
	       if (sscanf(b3,"%lf",&d1) != 1) return(3);
	       if (sscanf(b4,"%d,%d",&n1,&n2) != 2) return(4);
	       draw1_grid(d1,n1,n2,1,a);
	     }
	   else if (test_token("xz_grid",b2) == 1)
	     {
	       if (sscanf(b3,"%lf",&d1) != 1) return(3);
	       if (sscanf(b4,"%d,%d",&n1,&n2) != 2) return(4);
	       draw1_grid(d1,n1,n2,2,a);
	     }
	   else if (test_token("xy_grid",b2) == 1)
	     {
	       if (sscanf(b3,"%lf",&d1) != 1) return(3);
	       if (sscanf(b4,"%d,%d",&n1,&n2) != 2) return(4);
	       draw1_grid(d1,n1,n2,3,a);
	     }
	   else
	     {
		return(2);
		/*
		printf("Draw: invalid parameter!\n");
		 exit(1);*/
	     }
	 }
     }
   return(0);
}
		  
/************************************************************************
 * command_set
 ************************************************************************/

int command_set(PARAM *param, TASK *a)
{
   char *b2,*b3,*b4;
   double d1,d2,d3;
   int n1,n2,n3;

   if (param->count < 2) return(2);

   b2=param->v[1];
   if (test_token("null",b2) == 1) {
      set_null(a);
      return(0);
   }

   if (param->count < 3) return(3);

   b3=param->v[2];

   if (test_token("view",b2) == 1) {
      if (sscanf(b3,"%lf,%lf",&d1,&d2) != 2) return(3);
      a->view_alpha=d1;
      a->view_beta=d2;
      init_tmatrix(a);
   }

   else if (test_token("dist",b2) == 1) {
      if (sscanf(b3,"%lf",&d1) != 1) return(3);
      a->view_dist=d1;
   }

   else if (test_token("out_tol",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->out_tolerance=n1;
   }

   else if (test_token("shift",b2) == 1) {
      if (sscanf(b3,"%d,%d",&n1,&n2) != 2) return(3);
      a->image_center[0]=0.5*a->image_width + n1;
      a->image_center[1]=0.5*a->image_height + n2;
   }

   else if (test_token("zoom",b2) == 1) {
      if (sscanf(b3,"%lf",&d1) != 1) return(3);
      a->view_zoom=d1;
      init_tmatrix(a);
   }

   else if (test_token("using",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->object_pos[3]=n1;
   }

   else if (test_token("cb_pos",b2) == 1) {
      if (sscanf(b3,"%d,%d",&n1,&n2) != 2) return(3);
      if (n1 >= 0) {
	a->colorbar_pos[0]=n1;
      } else {
	a->colorbar_pos[0]=a->image_width+n1;
      }
      if (n2 >= 0) {
	a->colorbar_pos[1]=n2;
      } else {
	a->colorbar_pos[1]=a->image_height+n2;
      }
   }

   else if (test_token("cb_size",b2) == 1) {
      if (sscanf(b3,"%d,%d",&n1,&n2) != 2) return(3);
      a->colorbar_size[0]=n1;
      a->colorbar_size[1]=n2;
   }

   else if (test_token("cb_skip",b2) == 1) {
      if (sscanf(b3,"%d,%d",&n1,&n2) != 2) return(3);
      a->colorbar_skip[0]=n1;
      a->colorbar_skip[1]=n2;
   }

   else if (test_token("coord",b2) == 1) {
      if (sscanf(b3,"%d,%d,%d",&n1,&n2,&n3) != 3) return(3);
      a->object_pos[0]=n1;
      a->object_pos[1]=n2;
      a->object_pos[2]=n3;
   }

   else if (test_token("line_c",b2) == 1) {
      if (sscanf(b3,"%d,%d,%d",&n1,&n2,&n3) != 3) return(3);
      a->line_color[0]=n1;
      a->line_color[1]=n2;
      a->line_color[2]=n3;
   }

   else if (test_token("edge_c",b2) == 1) {
      if (sscanf(b3,"%d,%d,%d",&n1,&n2,&n3) != 3) return(3);
      a->edge_color[0]=n1;
      a->edge_color[1]=n2;
      a->edge_color[2]=n3;
   }

   else if (test_token("border_c",b2) == 1) {
      if (sscanf(b3,"%d,%d,%d",&n1,&n2,&n3) != 3) return(3);
      a->border_color[0]=n1;
      a->border_color[1]=n2;
      a->border_color[2]=n3;
   }

   else if (test_token("face_c",b2) == 1) {
      if (sscanf(b3,"%d,%d,%d",&n1,&n2,&n3) != 3) return(3);
      a->face_color[0]=n1;
      a->face_color[1]=n2;
      a->face_color[2]=n3;
      a->hidden_color[0]=n1;
      a->hidden_color[1]=n2;
      a->hidden_color[2]=n3;
   }

   else if (test_token("hidden_c",b2) == 1) {
      if (sscanf(b3,"%d,%d,%d",&n1,&n2,&n3) != 3) return(3);
      a->hidden_color[0]=n1;
      a->hidden_color[1]=n2;
      a->hidden_color[2]=n3;
   }

   else if (test_token("text_fg",b2) == 1) {
      if (sscanf(b3,"%d,%d,%d",&n1,&n2,&n3) != 3) return(3);
      a->text_color_fg[0]=n1;
      a->text_color_fg[1]=n2;
      a->text_color_fg[2]=n3;
   }

   else if (test_token("text_bg",b2) == 1) {
      if (sscanf(b3,"%d,%d,%d",&n1,&n2,&n3) != 3) return(3);
      a->text_color_bg[0]=n1;
      a->text_color_bg[1]=n2;
      a->text_color_bg[2]=n3;
   }

   else if (test_token("center",b2) == 1) {
      switch (b3[0])
	{
	 case 'l': a->text_center[0]=-1; break;
	 case 'c': a->text_center[0]=0; break;
	 case 'r': a->text_center[0]=1; break;
	 default: return(3);
	}
      switch (b3[1])
	{
	 case 't': a->text_center[1]=-1; break;
	 case 'c': a->text_center[1]=0; break;
	 case 'b': a->text_center[1]=1; break;
	 default: return(3);
	}
   }

   else if (test_token("format",b2) == 1) {
      set_axis_format(b3,a);
   }

   else if (test_token("app",b2) == 1) {
      set_appear(b3,a);
   }

   else if (test_token("box",b2) == 1) {
      if (sscanf(b3,"%lf,%lf,%lf",&d1,&d2,&d3) != 3) return(3);
      set_view_box(d1,d2,d3,a);
   }

   else if (test_token("auto_r",b2) == 1) {
      switch (b3[0])
	{
	 case 'x': a->view_auto_scale[0]=1; break;
	 case 'y': a->view_auto_scale[1]=1; break;
	 case 'z': a->view_auto_scale[2]=1; break;
	 default: return(3);
	}
   }

   else if (test_token("xrange",b2) == 1) {
      if (sscanf(b3,"%lf,%lf",&d1,&d2) != 2) return(3);
      a->view_start[0]=d1;
      a->view_end[0]=d2;
      a->view_auto_scale[0]=0;
   }

   else if (test_token("yrange",b2) == 1) {
      if (sscanf(b3,"%lf,%lf",&d1,&d2) != 2) return(3);
      a->view_start[1]=d1;
      a->view_end[1]=d2;
      a->view_auto_scale[1]=0;
   }

   else if (test_token("zrange",b2) == 1) {
      if (sscanf(b3,"%lf,%lf",&d1,&d2) != 2) return(3);
      a->view_start[2]=d1;
      a->view_end[2]=d2;
      a->view_auto_scale[2]=0;
   }

   else if (test_token("axis_ti",b2) == 1) {
      if (sscanf(b3,"%lf,%lf",&d1,&d2) != 2) return(3);
      a->axis_tick_len=d1;
      a->axis_tick_jump=d2;
   }

   else if (test_token("cont",b2) == 1) {
      if (sscanf(b3,"%lf,%lf,%lf",&d1,&d2,&d3) != 3) return(3);
      set_cont_levels(d1,d2,d3,a);
   }

   else if (test_token("light",b2) == 1) {
      if (param->count < 4) return(1);
      b4=param->v[3];
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      if (sscanf(b4,"%lf,%lf,%lf",&d1,&d2,&d3) != 3) return(4);
      if ((n1 >= 0) && (n1 < MAXLIGHTS)) {
	 a->light[n1].alpha=d1;
	 a->light[n1].beta=d2;
	 a->light[n1].power=d3;
      }
      else {
	 printf("Invalid light: %d\n",n1);
      }
   }

   else if (test_token("render_am",b2) == 1) {
      if (sscanf(b3,"%lf",&d1) != 1) return(3);
      a->render_ambient=d1;
   }

   else if (test_token("line_force",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->line_check_width=n1;
   }

   else if (test_token("line_w",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->line_width=n1;
   }

   else if (test_token("edge_w",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->edge_width=n1;
   }

   else if (test_token("border_w",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->border_width=n1;
   }

   else if (test_token("xticks",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->axis_ticks[0]=n1;
   }

   else if (test_token("yticks",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->axis_ticks[1]=n1;
   }

   else if (test_token("zticks",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->axis_ticks[2]=n1;
   }

   else if (test_token("cb_ticks",b2) == 1) {
      if (sscanf(b3,"%d",&n1) != 1) return(3);
      a->colorbar_ticks=n1;
   }

   else if (test_token("line_sh",b2) == 1) {
      YESNO(a->line_shading);
   }

   else if (test_token("line_dz",b2) == 1) {
      if (sscanf(b3,"%lf",&d1) != 1) return(3);
      a->line_dz=d1;
   }

   else if (test_token("edge_sh",b2) == 1) {
      YESNO(a->edge_shading);
   }

   else if (test_token("lut",b2) == 1) {
      load_lut(b3,a);
   }

   else if (test_token("font",b2) == 1) {
      load_font(b3,a);
   }

   else if (test_token("face_sh",b2) == 1) {
      switch (b3[0])
	{
	 case 'n': a->face_shading=0; break;
	 case 'y': a->face_shading=1; break;
	 case 'r': a->face_shading=2; break;
	 default: return(3);
	}
   }

   else if (test_token("obj_edge",b2) == 1) {
      YESNO(a->object_edges);
   }

   else if (test_token("obj_face",b2) == 1) {
      YESNO(a->object_faces);
   }

   else if (test_token("obj_cont",b2) == 1) {
      YESNO(a->object_cont);
   }

   else if (test_token("periodic_lut",b2) == 1) {
      YESNO(a->lut_periodic);
   }

   else if (test_token("grid_f",b2) == 1) {
      YESNO(a->grid_faces);
   }

   else if (test_token("grid_b",b2) == 1) {
      YESNO(a->grid_border);
   }

   else {
      return(2);
      /*
      printf("Set: invalid parameter!\n");
      exit(1);*/
   }
   return(0);
}
		  
/************************************************************************
 * print_syntax_error
 ************************************************************************/

static void print_syntax_error(PARAM *param, int m)
{
   int i,n;
   int len;

   m--;
   for (n=0; n < param->count; n++)
     {
	printf(" ");
	len=strlen(param->v[n]);
	if (param->flag[n] > 0) len += 2;
	if (n != m)
	  {
	     for (i=0; i < len; i++) printf(" ");
	  }
	else
	  {
	     for (i=0; i < len; i++) printf("^");
	     printf("\n");
	     break;
	  }
     }
   if (m >= param->count)
     {
	printf(" ^\n");
     }
   printf("Syntax error.\n");
}

/************************************************************************
 * mainloop
 ************************************************************************/

void mainloop(FILE *fp, TASK *a)
{
  PARAM param;
  char *b;
  char rbuf[MAXLINE];
  int k,m,n;

  while (1)
    {
      k=getline(fp,rbuf,MAXLINE);
      eval_param(&param, rbuf);
      if (param.count > 0)
	{
	  for (n=0; n < param.count; n++)
	    {
	       if (param.flag[n] == 1) {printf(" \"%s\"",param.v[n]);}
	       else if (param.flag[n] == 2) {printf(" [%s]",param.v[n]);}
	       else {printf(" %s",param.v[n]);}
	    }
	  printf("\n");
	  b=param.v[0];
	  m=0;
	  if (b[0] != '#') {
	    if (test_token("draw",b) == 1) {
	       m=command_draw(&param,a);
	    }
	    else if (test_token("text",b) == 1) {
	       m=command_text(&param,a);
	    }
	    else if (test_token("3dtext",b) == 1) {
	       m=command_3dtext(&param,a);
	    }
	    else if (test_token("xaxis",b) == 1) {
	       m=command_axis(&param,0,a);
	    }
	    else if (test_token("yaxis",b) == 1) {
	       m=command_axis(&param,1,a);
	    }
	    else if (test_token("zaxis",b) == 1) {
	       m=command_axis(&param,2,a);
	    }
	    else if (test_token("set",b) == 1) {
	       m=command_set(&param,a);
	    }
	    else if (test_token("back",b) == 1) {
	       m=background(&param,a);
	    } else {
	       m=1;
	    }
	    if (m > 0) {
	       print_syntax_error(&param,m);
	       exit(1);
	    }
	  }
	}
      if (k == 1) break;
    }
}

/************************************************************************
 * Inifile
 ************************************************************************/

static void chk_inifile(char *name, TASK *a)
{
   PARAM param;
   FILE *fp;
   char rbuf[MAXLINE];
   int k,n1,n2;

   fp=fopen(name,"r");
   if (fp == NULL) return;
   
   /*printf("Read: '%s'\n",name);*/
   while (1)
     {
	k=getline(fp,rbuf,MAXLINE);
	eval_param(&param, rbuf);
	if (param.count > 1)
	  {
	     if (test_token("font",param.v[0]) == 1)
	       {
		  strncpy(a->fontpath,param.v[1],MAXLEN-2);
	       }
	     else if (test_token("geom",param.v[0]) == 1)
	       {
		  if (sscanf(param.v[1],"%dx%d",&n1,&n2) == 2)
		    {
		       a->image_width=n1;
		       a->image_height=n2;
		    }
	       }
	     else if (test_token("line_jump",param.v[0]) == 1)
	       {
		  if (sscanf(param.v[1],"%d",&n1) == 1)
		    {
		       a->line_jump_max=n1;
		    }
	       }
	     else if (test_token("lut_max",param.v[0]) == 1)
	       {
		  if (sscanf(param.v[1],"%d",&n1) == 1)
		    {
		       a->lut_max_entries=n1;
		    }
	       }
	  }

	if (k == 1) break;
     }

   fclose(fp);
}

/************************************************************************
 * main
 ************************************************************************/

int main(int argc, char *argv[])
{
   TASK a;
   FILE *fp;
   char *pp;
   char *fname;
   char *home;
   char iname[MAXLINE];
   int res, kres;
   int len;
   int i;

   a.image_width=640;
   a.image_height=480;
   a.line_jump_max=1024;
   a.lut_max_entries=256;
   strcpy(a.ppmfile,ppmfile_name);
   strcpy(a.fontpath,fontpath_name);
   fname=NULL;

   home=getenv("HOME");
   len=strlen(home) + strlen(inifile) + 2;
   if (len < MAXLINE)
     {
	sprintf(iname,"%s/%s",home,inifile);
	chk_inifile(iname,&a);
     }
   
   for (i=1; i < argc; i++) {
      pp=argv[i];
      if (pp[0] == '-') {
	 res=kres=0;
	 switch(pp[1])
	   {
	    case 'g':
	      res=sscanf(pp+2,"%dx%d",&a.image_width,&a.image_height);
	      kres=2;
	      break;
	    case 'j':
	      res=sscanf(pp+2,"%d",&a.line_jump_max);
	      kres=1;
	      break;
	    case 'l':
	      res=sscanf(pp+2,"%d",&a.lut_max_entries);
	      kres=1;
	      break;
	    case 'o':
	      if (pp[2] != '\0')
	       {
		  strncpy(a.ppmfile,pp+2,MAXLEN-2);
	       }
	      break;
	    default:
	      kres=1;
	      break;
	   }
	 if (res != kres) {
	    fprintf(stderr,"Pixmax version 2.1, Klaus Ehrenfried (c) 1997,2002\n");
	    fprintf(stderr,"Usage: %s [options] [file]\n",argv[0]);
	    exit(1);
	 }
      }
      else {
	 fname=argv[i];
      }
   }

   if ((a.image_width < 2) || (a.image_height < 2))
     {
	printf("Invalid geometry %dx%d\n",a.image_width,a.image_height);
	exit(1);
     }
   if (a.line_jump_max < 2)
     {
	printf("Invalid line_jump_max %d\n",a.line_jump_max);
	exit(1);
     }
   if (a.lut_max_entries < 2)
     {
	printf("Invalid lut_max_entries %d\n",a.lut_max_entries);
	exit(1);
     }

   init_task(&a);

   if (fname != NULL) {
     fp=fopen(fname,"r");
     if (fp == NULL) {
       fprintf(stderr,"File '%s' not found\n",fname);
       exit(1);
     }
   }
   else {
     fp=stdin;
   }

   mainloop(fp,&a);

   if (fname != NULL) {
     fclose(fp);
   }

   write_ppm(&a);

   return(0);
}
