/***********************************************************************
 * main.h
 *  
 * Copyright (C) 1997-2002 Klaus Ehrenfried
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * 
 * Klaus Ehrenfried 26-Mar-97
 * 
 * 7-Mar-2000 added hidden_color K.E.
 ***********************************************************************/

#define UBYTE unsigned char
#define MAXLEN 1024
#define MAXLIGHTS 3
#define MAXCONTLEV 1024
#define FMTLEN 20

#ifndef PI
#define PI M_PI
#endif

typedef struct light_struct
{
   double d[3];
   double alpha,beta;
   double power;
   int flag;
} LIGHT;

typedef struct lut_entry_struct {
   double w1,w2;
   double step;
   UBYTE *table;
   int max;
} LUT_ENTRY;

typedef struct rpoint_struct {
   int i,j;
   double z;
   double w;
} RPOINT;

typedef struct point_struct {
   double x,y,z;
   double w;
} POINT;

typedef struct face_struct {
   int id[4];
} FACE;

typedef struct pixel_struct {
   UBYTE r,g,b;
   UBYTE hit;
   double z;
} PIXEL;

typedef struct face_object_struct
{
   POINT *pv;
   FACE *pf;
   int v_count;
   int f_count;
} FACE_OBJECT;

typedef struct mesh_object_struct {
  POINT *pv;
  int *id;
  int len;
  int m;
  int n;
  int rect_flag;
} MESH_OBJECT;

typedef struct task_struct {
   PIXEL *pels;
   int image_width;
   int image_height;
   int image_len;
   double image_center[2];

   UBYTE *font_pels;
   int font_width;
   int font_height;
   int font_scanline;
   int font_mem_len;

   int text_center[2];
   int text_color_bg[3];
   int text_color_fg[3];
   POINT text1_point;

   int colorbar_pos[2];
   int colorbar_size[2];
   int colorbar_skip[2];
   int colorbar_ticks;

   int axis_ticks[3];
   double axis_tick_len;
   double axis_tick_jump;
   char axis_format[FMTLEN];

   LUT_ENTRY *lut;
   int lut_max_entries;
   int lut_count;
   int lut_periodic;
   int lut_auto;
   double lut_wmin;
   double lut_wmax;
   double lut_dw;

   LIGHT light[MAXLIGHTS];
   double render_ambient;
   double render_color[3];
   double render_hidden_color[3];

   double cont_value[MAXCONTLEV];
   int cont_levels;

   int *line_jump;
   int line_jump_len;
   int line_jump_max;
   int line_pixel_shift;
   int line_check_width;
   int out_tolerance;
   double line_dz;
   /*double line_check_dz;*/

   int line_width;
   int edge_width;
   int border_width;

   int line_shading;
   int face_shading;
   int edge_shading;
   int draw_clu_flag;
   int draw_shading_flag;

   double view_alpha;
   double view_beta;
   double view_dist;
   double view_zoom;
   double view_box[3];
   double view_start[3];
   double view_end[3];
   double view_origin[3];
   double view_scale[3];
   int view_auto_scale[3];

   double tmatrix_h[3];
   double tmatrix_v[3];
   double tmatrix_z[3];

   int object_cont;
   int object_edges;
   int object_faces;
   int object_pos[4];

   int line_color[3];
   int face_color[3];
   int hidden_color[3];
   int edge_color[3];
   int border_color[3];
   int work_color[3];

   int grid_faces;
   int grid_border;

   char ppmfile[MAXLEN];
   char fontpath[MAXLEN];
} TASK;

