/************************************************************************
 * readoff.c
 *
 * Copyright (C) 1997-2002 Klaus Ehrenfried
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Klaus Ehrenfried 26-MAR-1997
 * 
 * 17-May-99 pick_double: jetzt auch npos == 0 zulaessig
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "main.h"
#include "object.h"

#define MAXLINE 512

/************************************************************************
 *   Lese Zeile
 ***********************************************************************/

static int getline(FILE *fp, char s[], int len)
{
  int i,c;

  for (i=0; i < len-1; i++)
    {
      c=getc(fp);
      if (c == EOF)
	{
	  s[i]='\0';
	  return(1);
	}
      if (c == '\n')
	{
	  s[i]='\0';
	  return(0);
	}
      if (c == '\t') c=' ';
      s[i]=c;
    }
  s[len-1]='\0';
  return(0);
}

/************************************************************************
 *   Suche ersten Character von n-ter Position im Record
 ***********************************************************************/

static char *getstart(char *s, int len, int npos)
{
  int count_down;
  int active;
  int i;

  count_down=npos-1;
  active=1;

  for (i=0; i < len-1; i++) {
    if (*s == '\0') break;

    if (active == 1) {
      if (*s != ' ') {
      	if (count_down <= 0) break;
	count_down--;
	active=0;
      }
    }
    else {
      if (*s == ' ') active=1;
    }
    s++;
  }

  return(s);
}

/************************************************************************
 *   Lese n-te Position als double
 ***********************************************************************/

int pick_double(char *rbuf, int len, int npos, double *w)
{
   char *b;
   int k;

   if (npos == 0) {
      *w = 0.0;
      return(1);
   }
   b=getstart(rbuf,len,npos);
   k=sscanf(b,"%lf",w);

   if (k == 1) return(1);

   return(0);
}

/************************************************************************
 *   Lese OFF datei
 ***********************************************************************/

int read_off(char fname[], FACE_OBJECT *a, int pos[4])
{
   FILE *fp;
   POINT *pv;
   FACE *pf;
   char rbuf[MAXLINE];
   char *b;
   double h[4];
   int off_flag;
   int k,ii,j;
   int nv,n1,n2,n3,n4;
   int iv,ie;

   fp=fopen(fname,"r");
   if (fp == NULL) {
      printf("File '%s' not found\n",fname);
      return(1);
   }

   a->v_count = 0;
   a->f_count = 0;
   off_flag = 0;

   while (off_flag == 0)
     {
	k=getline(fp,rbuf,MAXLINE);
	b=getstart(rbuf,MAXLINE,1);
	/*printf("DEBUG: '%s'\n",b);*/
	if ((*b != '#') && (*b != '\0'))
	  {
	     ii=sscanf(b,"%d %d",&n1,&n2);
	     if (ii == 2)
	       {
		  a->v_count = n1;
		  a->f_count = n2;
		  off_flag = 1;
	       }
	  }
	if (k == 1) break;
     }
   
   printf("points=%d elements=%d\n",a->v_count,a->f_count);
   if (off_flag != 1)
     {
	printf("No header in '%s'\n",fname);
	return(1);
     }

   pv=(POINT *)calloc(a->v_count, sizeof(POINT));
   if (pv == NULL)
     {
	printf("No memory - v_count: %d\n",a->v_count);
	return(1);
     }

   pf=(FACE *)calloc(a->f_count, sizeof(FACE));
   if (pf == NULL)
     {
	printf("No memory - f_count: %d\n",a->f_count);
	return(1);
     }

   a->pv=pv;
   a->pf=pf;

   iv=ie=0;

   while (iv < a->v_count)
     {
	k=getline(fp,rbuf,MAXLINE);
	b=getstart(rbuf,MAXLINE,1);
	if ((*b != '#') && (*b != '\0'))
	  {
	    for (j=0; j < 4; j++) {
	      ii=pick_double(rbuf, MAXLINE, pos[j], h+j);
	      if (ii != 1) {
		fprintf(stderr,"Error reading column %d of vertex %d:\n",
			pos[j],iv);
		fprintf(stderr,"File:   '%s'\n",fname);
		fprintf(stderr,"Record: '%s'\n",rbuf);
		return(1);
	      }
	    }
	    pv->x=h[0];
	    pv->y=h[1];
	    pv->z=h[2];
	    pv->w=h[3];
	    pv++;
	    iv++;
	  }
	if (k == 1) break;
     }
 
   if (iv != a->v_count)
     {
	printf("Not enough vertices in file (%d)\n",iv);
	return(1);
     }

   while (ie < a->f_count) {
      k=getline(fp,rbuf,MAXLINE);
      b=getstart(rbuf,MAXLINE,1);
      if (*b != '#') {
	 
	 ii=sscanf(b,"%d",&nv);
	 if (ii > 0) {
	    if (nv == 3)
	      {
		 ii=sscanf(b,"%d %d %d %d",&nv,&n1,&n2,&n3);
		 n4=n3;
	      }
	    else if (nv == 4)
	      {
		 ii=sscanf(b,"%d %d %d %d %d",&nv,&n1,&n2,&n3,&n4);
	      }
	    else {
	       fprintf(stderr,"%d vertices in element %d\n",nv,ie);
	       fprintf(stderr,"Only 3 or 4 supported, sorry.\n");
	       return(1);
	    }
	    if (ii != (nv+1))
	      {
		 fprintf(stderr,"Error in element %d\n",ie);
		 fprintf(stderr,"'%s'\n",b);
		 return(1);
	      }
	    pf->id[0]=n1;
	    pf->id[1]=n2;
	    pf->id[2]=n3;
	    pf->id[3]=n4;
	    pf++;
	    ie++;
	 }
      }
      if (k == 1) break;
   }

   fclose(fp);
   return(0);
}

/************************************************************************
 *   Lese mesh file
 ***********************************************************************/

int read_mesh(char fname[], MESH_OBJECT *mesh, int pos[4])
{
   FILE *fp;
   POINT *pv;
   double h[4];
   char rbuf[MAXLINE];
   char *b;
   int *id;
   int space_flag;
   int iv,ic;
   int k,ii,j;
   int n,nmax;

   fp=fopen(fname,"r");
   if (fp == NULL) {
      printf("File '%s' not found\n",fname);
      return(1);
   }

   /* -- scan file -- */

   iv=ic=0;
   space_flag=1;
   while (1)
     {
	k=getline(fp,rbuf,MAXLINE);
	b=getstart(rbuf,MAXLINE,1);
	if (*b != '#')
	  {
	    if (*b == '\0') {
	      space_flag=1;
	    }
	    else {
	      iv++;
	      if (space_flag == 1) ic++;
	      space_flag=0;
	    }
	  }
	if (k == 1) break;
     }
   
   printf("read_mesh - points=%d  columns=%d\n",iv,ic);

   if (ic == 0) {
     printf("File contains no points\n");
     return(1);
   }

   mesh->len=iv;
   mesh->m=ic;

   pv=(POINT *)calloc(mesh->len, sizeof(POINT));
   if (pv == NULL)
     {
	printf("No memory - %d points\n",mesh->len);
	return(1);
     }

   id=(int *)calloc(mesh->m+1, sizeof(int));
   if (id == NULL)
     {
	printf("No memory - %d integers\n",mesh->m);
	return(1);
     }

   mesh->pv=pv;
   mesh->id=id;

   rewind(fp);

   iv=ic=0;
   space_flag=1;

   while (iv < mesh->len)
     {
       k=getline(fp,rbuf,MAXLINE);
       b=getstart(rbuf,MAXLINE,1);
       if (*b != '#')
	 {
	   if (*b == '\0')
	     {
	       space_flag=1;
	     }
	   else
	     {
	       if (space_flag == 1)
		 {
		   if (ic >= mesh->m)
		     {
		       printf("Corrupted file\n");
		       exit(1);
		     }
		   id[ic]=iv;
		   ic++;
		 }
	       space_flag=0;
		    
	       for (j=0; j < 4; j++) {
		 ii=pick_double(rbuf, MAXLINE, pos[j], h+j);
		 if (ii != 1) {
		   fprintf(stderr,"Error reading column %d of vertex %d:\n",
			   pos[j],iv);
		   fprintf(stderr,"File:   '%s'\n",fname);
		   fprintf(stderr,"Record: '%s'\n",rbuf);
		   return(1);
		 }
	       }
	       pv->x=h[0];
	       pv->y=h[1];
	       pv->z=h[2];
	       pv->w=h[3];
	       pv++;
	       iv++;
	     }
	 }
       if (k == 1) break;
     }

   if ((iv != mesh->len) || (ic != mesh->m))
     {
       printf("Impossible change - now points=%d columns=%d\n",iv,ic);
       return(1);
     }

   fclose(fp);

   id[mesh->m]=mesh->len;

   nmax=0;
   for (j=0; j < mesh->m; j++)
     {
       n=id[j+1] - id[j];
       if (n > nmax) nmax=n;
     }

   mesh->n=nmax;
   if ((mesh->n * mesh->m) == mesh->len) {
     mesh->rect_flag = 1;
     printf("read_mesh: %dx%d\n", mesh->m, mesh->n);
   }
   else {
     mesh->rect_flag = 0;
     printf("read_mesh: floating column len (max=%d)\n", mesh->n);
   }

   return(0);
}
