/***********************************************************************
 * text.c
 * 
 * Copyright (C) 1997-2002 Klaus Ehrenfried
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * 
 * Klaus Ehrenfried 31-MAR-1997
 * 28-Dec-2001 load_font: switched to bdf file format
 * 7-Jan-2002 draw0_colorbar: added border
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#define MAXLINE 512
#define MAXTEXTLEN 1024
#define TINY 1e-10

#include "main.h"
#include "draw3d.h"
#include "text.h"

/************************************************************************
 *   Lese Zeile
 ***********************************************************************/

static int test_string(char s1[], char s2[])
{
   int i;
   i=0;
   while (s2[i] != '\0') {
      if (s1[i] != s2[i]) return(0);
      i++;
   }
   return(1);
}

/************************************************************************
 *   Lese Zeile
 ***********************************************************************/

static int getline(FILE *fp, char s[], int len)
{
  int i,c;

  for (i=0; i < len-1; i++)
    {
      c=getc(fp);
      if (c == EOF)
        {
          s[i]='\0';
          return(1);
        }
      if (c == '\n')
        {
          s[i]='\0';
          return(0);
        }
      s[i]=c;
    }
  s[len-1]='\0';
  return(0);
}

/************************************************************************
 *
 ***********************************************************************/

static const UBYTE bmask[8]={128,64,32,16,8,4,2,1};

/************************************************************************
 *
 ***********************************************************************/

static int test_font_pel(int nchar, int i0, int j0, TASK *a)
{
   int m,n;
   UBYTE *b;
   
   m = i0 % 8;
   n = (nchar * a->font_height + j0) * a->font_scanline + (i0/8);

   if (n >= a->font_mem_len) {
      printf("Error test_font_pel\n");
      exit(1);
   }
   
   b = a->font_pels + n;

   return(*b & bmask[m]);
}

/************************************************************************
 *
 ***********************************************************************/

static void set_font_pel(int nchar, int row, char *s, TASK *a)
{
   int n;
   int w;
   UBYTE *b;
   char hb[3];

   if ((nchar > 255) || ( row > a->font_height)) {
      printf("Error set_font_pel\n");
      exit(1);
   }
   
   b = a->font_pels + ((nchar * a->font_height + row) * a->font_scanline);

   hb[2]='\0';
   w=0;
   for (n=0; n < a->font_scanline; n++) {
      hb[0]=s[0];
      hb[1]=s[1];
      if (sscanf(hb,"%X", &w) != 1) {
	 printf("Error set_font_pel\n");
	 printf("%s\n", s);
	 exit(1);
      }
      b[n] = w;
      s += 2;
   }
}

/********************************************************************
 * Bearbeite backslash-Chars
 ********************************************************************/

static int backslash(int *s, char *text, int len)
{
  int i,n,flag;

  flag=0;
  i=0;
  for (n=0; n < len; n++)
    {
      if (text[n] == '\\')
	{
	  if (flag == 1)
	    {
	      s[i++] = text[n];
	      flag=0;
	    }
	  else
	    {
	      flag=1;
	    }
	}
      else
	{
	  if ((flag == 1) && (*s > 0))
	    {
	      s[i++] = (int)text[n] + 128;
	    }
	  else
	    {
	      s[i++] = text[n];
	    }
	  flag=0;
	}
    }
  return(i);
}

/********************************************************************
 * draw0_text
 ********************************************************************/

void draw0_text(RPOINT *rp, char *text, TASK *a)
{
  int *c;
  int len;
  int w,h;
  int js,is;
  int i,j,k,m,n;
  int s[MAXTEXTLEN];

  if (a->font_pels == NULL)
    {
      printf("No font loaded\n");
      return;
    }

  /*printf("Text: '%s'\n",s);*/
  len=strlen(text);
  if (len > MAXTEXTLEN)
    {
      printf("Text too long (truncated)\n");
      len=MAXTEXTLEN;
    }
  if (len == 0)
    {
      printf("Empty string\n");
      return;
    }

  len=backslash(s,text,len);

  w=len*a->font_width;
  h=a->font_height;

  switch (a->text_center[0])
    {
    case 0: is=rp->i - w/2; break;
    case 1: is=rp->i -w; break;
    default: is=rp->i; break;
    }

  switch (a->text_center[1])
    {
    case 0: js=rp->j - h/2; break;
    case 1: js=rp->j -h; break;
    default: js=rp->j; break;
    }

  /*
     printf("len=%d  w=%d h=%d\n",len,w,h);
     printf("is=%d  js=%d\n",is,js);
     */

   j=js;
   for (m=0; m < a->font_height; m++)
     {
	i=is;
	for (n=0; n < len; n++)
	  {
	     for (k=0; k < a->font_width; k++)
	       {
		  if (test_font_pel(s[n],k,m,a) == 0)
		    {
		       c=a->text_color_bg;
		    }
		  else
		    {
		       c=a->text_color_fg;
		    }
		  set0_pixel(i,j,rp->z,c,a);
		  i++;
	       }
	  }
	j++;
     }
}

/********************************************************************
 * draw1_text
 ********************************************************************/

void draw1_text(POINT *p1, char *s, TASK *a)
{
  RPOINT rp;

  trans10(p1, &rp, 0, a);
  draw0_text(&rp,s,a);
}

/********************************************************************
 * draw1_axis
 ********************************************************************/

void draw1_axis(int dir, double h[2], TASK *a)
{
   POINT p1,p2,p3;
   int i,j,n;
   char record[MAXTEXTLEN];
   double dw,dr,w,r;

   if ((dir < 0) || (dir > 2)) return;

   n=a->axis_ticks[dir];

   if (n < 2) n=2;

   dw=(a->view_end[dir] - a->view_start[dir])/(double)(n-1);
   dr=2.0/(double)(n-1);

   p1.w=0.0;
   p2.w=0.0;

   if (dir == 0)
     {
	p1.x=-1.0;
	p1.y=h[0];
	p1.z=h[1];
	p2.x=+1.0;
	p2.y=h[0];
	p2.z=h[1];
     }
   else if (dir == 1)
     {
	p1.x=h[0];
	p1.y=-1.0;
	p1.z=h[1];
	p2.x=h[0];
	p2.y=+1.0;
	p2.z=h[1];
     }
   else
     {
	p1.x=h[0];
	p1.y=h[1];
	p1.z=-1.0;
	p2.x=h[0];
	p2.y=h[1];
	p2.z=+1.0;
     }

   for (j=0; j < 3; j++) {
      a->work_color[j] = a->line_color[j];
   }
   a->draw_clu_flag=0;
   set_line_width(a->line_width,a);

   p1.x *= a->view_box[0];
   p1.y *= a->view_box[1];
   p1.z *= a->view_box[2];

   p2.x *= a->view_box[0];
   p2.y *= a->view_box[1];
   p2.z *= a->view_box[2];

   draw1_line(&p1, &p2, a);

   p2.x = p1.x * a->axis_tick_len;
   p2.y = p1.y * a->axis_tick_len;
   p2.z = p1.z * a->axis_tick_len;
   p2.w = p1.w;

   p3.x = p1.x * a->axis_tick_jump;
   p3.y = p1.y * a->axis_tick_jump;
   p3.z = p1.z * a->axis_tick_jump;
   p3.w = p1.w;

   for (i=0; i < n; i++)
     {
	w=a->view_start[dir] + i * dw;
	r=(-1.0 + i * dr) * a->view_box[dir];
	if ((w < (TINY*dw)) && (w > (-TINY*dw))) w=0.0;
	sprintf(record,a->axis_format,w);

	switch (dir)
	  {
	   case 0: p1.x = p2.x = p3.x = r; break;
	   case 1: p1.y = p2.y = p3.y = r; break;
	   default: p1.z = p2.z = p3.z = r; break;
	  }

	/*printf("axis: '%s' (%f,%f,%f)\n",record,p.x,p.y,p.z);*/
	draw1_line(&p1, &p2, a);
	draw1_text(&p3, record, a);
     }
}

/********************************************************************
 * draw0_colorbar
 ********************************************************************/

void draw0_colorbar(double h0, double h1, TASK *a)
{
   RPOINT p1,p2,p3;
   char record[MAXTEXTLEN];
   int i1,i2;
   int j1,j2;
   double w11,w12,w21,w22;
   double step,dw,dr,w;
   int dir;
   int m,n;

   p3.z=p2.z=p1.z=0.0;
   p3.w=p2.w=p1.w=0.0;

   i1=a->colorbar_pos[0];
   i2=i1+a->colorbar_size[0];

   j1=a->colorbar_pos[1];
   j2=j1+a->colorbar_size[1];

   if (a->colorbar_size[0] > a->colorbar_size[1])
     {
	w11=w21=h0;
	w12=w22=h1;
	a->text_center[0]=0;
	a->text_center[1]=-1;
	step=i2-i1;
	dir=1;
     }
   else
     {
	w11=w12=h0;
	w21=w22=h1;
	a->text_center[0]=1;
	a->text_center[1]=0;
	step=j2-j1;
	dir=2;
     }

   fill0_trapezoid_clu
     (j1,i1,0.0,w11,i2,0.0,w12,j2,i1,0.0,w21,i2,0.0,w22,a);

   
   /* neu */
   n=a->colorbar_ticks;

   if (n > 0) {
      set_line_width(a->line_width,a);

      p1.i=i1; p1.j=j1;
      p2.i=i2; p2.j=j1;
      draw0_line_const(&p1, &p2, a->line_color, a);
      p1.i=i2; p1.j=j2;
      draw0_line_const(&p1, &p2, a->line_color, a);
      p2.i=i1; p2.j=j2;
      draw0_line_const(&p1, &p2, a->line_color, a);
      p1.i=i1; p1.j=j1;
      draw0_line_const(&p1, &p2, a->line_color, a);
   } else {
      n = -n;
   }
   if (n < 2) n=2;

   dw=(h1 - h0)/(double)(n-1);
   dr=step/(double)(n-1);

   set_line_width(a->line_width,a);

   for (m=0; m < n; m++)
     {
	w=h0 + m * dw;
	if ((w < (TINY*dw)) && (w > (-TINY*dw))) w=0.0;
	sprintf(record,a->axis_format,w);

	if (dir == 1)
	  {
	     p1.i=i1 + m * dr;
	     p1.j=j2;
	     p3.i=p2.i=p1.i;
	     p2.j=j2+a->colorbar_skip[0];
	     p3.j=j2+a->colorbar_skip[1];
	  }
	else
	  {
	     p1.i=i1;
	     p1.j=j1 + m *dr;
	     p3.j=p2.j=p1.j;
	     p2.i=i1-a->colorbar_skip[0];
	     p3.i=i1-a->colorbar_skip[1];
	  }
	draw0_line_const(&p1, &p2, a->line_color, a);
	draw0_text(&p3, record, a);
     }

}

/************************************************************************
 *   Lese font (bdf) datei
 ***********************************************************************/

int load_font(char fname[], TASK *a)
{
   FILE *fp;
   char rbuf[MAXLINE];
   char hbuf[MAXLINE];
   char font_file[MAXLEN];
   int i;
   int res;
   int f_width, f_height;
   int len;
   int nchar;
   int row;
   int bit_flag;

   len=strlen(a->fontpath) + strlen(fname) + 6;
   if (len > MAXLEN)
     {
	printf("File name too long: '%s/%s'\n",a->fontpath,fname);
	exit(1);
     }

   sprintf(font_file,"%s/%s",a->fontpath,fname);

   fp=fopen(font_file,"r");
   if (fp == NULL) {
      printf("File '%s' not found\n",font_file);
      exit(1);
   }

   // -- GET FONT DIMENSIONS --

   res=0;
   f_width=0;
   f_height=0;

   while (res == 0) {
      res = getline(fp, rbuf, MAXLINE);
      if (test_string(rbuf, "FONTBOUNDINGBOX")) {
	 //printf("%s\n", rbuf);
	 if (sscanf(rbuf, "%s %d %d", hbuf, &f_width, &f_height) != 3) {
	    printf("Error in FONT file\n");
	    exit(1);
	 }
      }
   }
   
   printf("Font size: %dx%d\n", f_width, f_height);

   a->font_width = f_width;
   a->font_height = f_height;

   // --- allocate memory ---
   
   a->font_scanline = f_width/8;
   if ((f_width % 8) > 0)
	   a->font_scanline++;

   if ((a->font_scanline * 2 + 1) >= MAXLINE) {
      printf("font too wide: %d\n", f_width);
      exit(1);
   }

   a->font_mem_len = 256*f_height*a->font_scanline;

   //printf("len: %d %d\n", a->font_mem_len, a->font_scanline);

   a->font_pels=(UBYTE *)malloc(a->font_mem_len);

   if (a->font_pels == NULL)
     {
        printf("No memory for font\n");
        exit(1);
     }

   for (i=0; i < len; i++) a->font_pels[i]=0;
   
   // --- get bitmaps ---

   nchar=0;
   res=0;
   row=0;
   bit_flag=0;
   rewind(fp);

   while (res == 0) {
      res = getline(fp, rbuf, MAXLINE);
      if (test_string(rbuf, "ENCODING")) {
	 //printf("%s\n", rbuf);
	 if (sscanf(rbuf, "%s %d", hbuf, &nchar) != 2) {
	    printf("Error reading font file\n");
	    exit(1);
	 }
	 nchar = nchar % 256;
      }
      else if (test_string(rbuf, "BITMAP")) {
	 //printf("%s\n", rbuf);
	 row=0;
	 bit_flag=1;
      }
      else if (test_string(rbuf, "ENDCHAR")) {
	 //printf("%s\n", rbuf);
	 bit_flag=0;
      } else if (bit_flag == 1) {
	 if (row > a->font_height) {
	    printf("Invalid row %d (char=%d)\n", row, nchar);
	    exit(1);
	 }
	 set_font_pel(nchar, row, rbuf, a);
	 row++;
      }
   }
   
   fclose(fp);

   return(0);
}

/* --- FIN --- */

